package frc.robot.subsystems;

import com.revrobotics.RelativeEncoder;
import com.revrobotics.spark.ClosedLoopSlot;
import com.revrobotics.spark.SparkClosedLoopController;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.ControlType;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;
import com.revrobotics.spark.config.SparkMaxConfig;
import com.revrobotics.spark.config.SparkBaseConfig.IdleMode;

import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.MotorIDConstants;

public class ClosedElevator extends SubsystemBase {
    //Encoder constants
    private static final double kPositionConversionFactor = 0;
    private static final double kVelocityConversionFactor = 0;
    //PID constants
    private static final double kP = 0;
    private static final double kI = 0;
    private static final double kD = 0;
    private static final double kFF = 0;
    //MAXMotion constants
    private static final double kMaxAcceleration = 0;
    private static final double kMaxVelocity = 0;
    private static final double kPIDTolerance = 0;
    //Homing constants
    private static final double kHomingLocation = 0;
    private static final double kHomingVoltage = 0;
    private static final double kHomingDelay = 0.1;
    private static final double kHomingVelocityLimit = 0;

    //Motor Config constants
    private static final SparkMaxConfig motorConfig = new SparkMaxConfig();

    static {
        motorConfig
            .idleMode(IdleMode.kBrake);

        motorConfig.encoder
            .positionConversionFactor(kPositionConversionFactor)
            .velocityConversionFactor(kVelocityConversionFactor);
        
        motorConfig.closedLoop
            .pidf(
                kP, 
                kI, 
                kD, 
                kFF, 
                ClosedLoopSlot.kSlot0
            );

        motorConfig.closedLoop.maxMotion
            .maxAcceleration(kMaxAcceleration, ClosedLoopSlot.kSlot0)
            .maxVelocity(kMaxVelocity, ClosedLoopSlot.kSlot0)
            .allowedClosedLoopError(kPIDTolerance, ClosedLoopSlot.kSlot0);
    }

    //----------------------------------------------------------------------------------------------------//

    private double goal;

    private RelativeEncoder encoder;

    private SparkMax motor1;

    private SparkClosedLoopController controller;

    public ClosedElevator() {
        motor1 = new SparkMax(
            MotorIDConstants.kElevatorMotor1ID, 
            MotorType.kBrushless
        );

        motor1.configure(
            motorConfig, 
            ResetMode.kResetSafeParameters,
            PersistMode.kPersistParameters
        );

        controller = motor1.getClosedLoopController();

        encoder = motor1.getEncoder();
    }

    /**
     * Returns whether or not the encoder position is within tolerance of the goal
     * 
     * @return True if within tolerance
     */
    public boolean atGoal() {
        return encoder.getPosition() >= goal - kPIDTolerance && encoder.getPosition() <= goal + kPIDTolerance;
    }

    /**
     * Sets the reference goal to the goal parameter
     * 
     * @param goal The desired goal
     * @return Command that provides power to the motor based on the closed loop controller's calculations
     */
    public Command goToPosition(double goal) {
        this.goal = goal;

        return run(() -> {
            controller.setReference(goal, ControlType.kMAXMotionPositionControl);
        });
    }

    /**
     * Runs the elevator to a set position slightly above home, then applies a constant voltage until velocity is equal to zero.
     * This then rezeros the encoder
     * 
     * @return Command that achieves a proper rezeroing position
     */
    public Command goToHome() {
        return run(() -> {
            controller.setReference(
                kHomingLocation, 
                ControlType.kMAXMotionPositionControl
            );
        }).until(() -> atGoal()).andThen(
            run(() -> {
                controller.setReference(kHomingVoltage, ControlType.kVoltage);
            }).withTimeout(kHomingDelay).andThen(
                run(() -> {
                    controller.setReference(kHomingVoltage, ControlType.kVoltage);
                })
            ).until(() -> encoder.getVelocity() <= kHomingVelocityLimit)
        ).finallyDo((interrupted) -> {
            if (!interrupted) {
                encoder.setPosition(0);
                goToPosition(0);
            }
        });
    }
}